#define _GNU_SOURCE
#include <pthread.h>
#include <stdio.h>
#include <string.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <stddef.h>
#include <stdlib.h>
#include <unistd.h>
#include <asm/kvm_para.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/syscall.h>
#include <linux/kvm.h>
#include <linux/userfaultfd.h>

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

struct idt_entry {
  unsigned long a;
  unsigned long b;
};

extern char guest_code_begin[];
extern char guest_code_end[];
extern char guest_code_vmmcall[];
asm(
"guest_code_begin:\n"
"movq $3, %rax\n"
"guest_code_vmmcall:\n"
"vmmcall\n"
"guest_code_end:\n"
);

struct guest_memory {
  /* 0x0000 */ unsigned long pagetable1[512];
  /* 0x1000 */ unsigned long pagetable2[512];
  /* 0x2000 */ unsigned long pagetable3[512];
  /* 0x3000 */ unsigned long pagetable4[512];
  /* 0x4000 */ unsigned char guest_code[4096];
  union {
    struct kvm_steal_time steal_time;
    unsigned char pad[4096];
  };
} __attribute__((aligned(4096)));
struct guest_memory gmem;
#define GPHYS(x) offsetof(struct guest_memory, x)

static void create_region(int vm, int slot, unsigned long guest_phys, unsigned long host_addr, unsigned long size) {
  struct kvm_userspace_memory_region region = {
    .slot = slot,
    .guest_phys_addr = guest_phys,
    .memory_size = size,
    .userspace_addr = host_addr
  };
  if (ioctl(vm, KVM_SET_USER_MEMORY_REGION, &region))
    err(1, "set region %d", slot);
}

static void wrmsr(int vcpu, unsigned int msr, unsigned long value) {
  struct kvm_msrs *arg = alloca(offsetof(struct kvm_msrs, entries) + sizeof(struct kvm_msr_entry));
  memset(arg, 0, offsetof(struct kvm_msrs, entries) + sizeof(struct kvm_msr_entry));
  arg->nmsrs = 1;
  arg->entries[0].index = msr;
  arg->entries[0].data = value;
  errno = 0;
  int ret = ioctl(vcpu, KVM_SET_MSRS, arg);
  if (ret != 1)
    err(1, "set MSR 0x%x returned %d", msr, ret);
}

static void *thread_fn(void *dummy) {
  while (1) {
    printf("[periodic check] kvm_steal_time: version=0x%x preempted=0x%hhx\n", gmem.steal_time.version, gmem.steal_time.preempted);
    sleep(1);
  }
}

int main(void) {
  gmem.pagetable1[0] = GPHYS(pagetable2) | 7;
  gmem.pagetable2[0] = GPHYS(pagetable3) | 7;
  gmem.pagetable3[0] = GPHYS(pagetable4) | 7;
  gmem.pagetable4[0] = GPHYS(guest_code) | 7;
  memset(gmem.guest_code, 0xcc, 0x1000);
  memcpy(gmem.guest_code, guest_code_begin, guest_code_end - guest_code_begin);

  int uffd = SYSCHK(syscall(SYS_userfaultfd, 0));
  struct uffdio_api api = { .api = UFFD_API, .features = 0 };
  SYSCHK(ioctl(uffd, UFFDIO_API, &api));
  struct uffdio_range trapped = {.start = (unsigned long)gmem.guest_code, .len = 0x1000};
  struct uffdio_register reg = {.mode = UFFDIO_REGISTER_MODE_WP, .range = trapped };
  SYSCHK(ioctl(uffd, UFFDIO_REGISTER, &reg));
  struct uffdio_writeprotect wrprot = { .range = {.start = (unsigned long)gmem.guest_code, .len = 0x1000}, .mode = UFFDIO_WRITEPROTECT_MODE_WP };
  SYSCHK(ioctl(uffd, UFFDIO_WRITEPROTECT, &wrprot));

  int kvm = open("/dev/kvm", O_RDWR);
  if (kvm == -1)
    err(1, "open kvm");
  int mmap_size = ioctl(kvm, KVM_GET_VCPU_MMAP_SIZE, 0);
  if (mmap_size == -1)
    err(1, "KVM_GET_VCPU_MMAP_SIZE");

  int vm = ioctl(kvm, KVM_CREATE_VM, 0);
  if (vm == -1)
    err(1, "create vm");
  if (ioctl(vm, KVM_SET_TSS_ADDR, 0x10000000UL))
    err(1, "KVM_SET_TSS_ADDR");
  create_region(vm, 0, 0x0, (unsigned long)&gmem, sizeof(gmem));

  int vcpu = ioctl(vm, KVM_CREATE_VCPU, 0);
  if (vcpu == -1)
    err(1, "create vcpu");
  struct kvm_run *vcpu_state = mmap(NULL, mmap_size, PROT_READ|PROT_WRITE, MAP_SHARED, vcpu, 0);
  if (vcpu_state == MAP_FAILED)
    err(1, "mmap vcpu");

  struct kvm_sregs sregs;
  if (ioctl(vcpu, KVM_GET_SREGS, &sregs))
    err(1, "KVM_GET_SREGS");
  sregs.cs.selector = 0;
  sregs.cs.base = 0;
  sregs.cr0 = (/*PE*/1<<0) | (/*WP*/1<<16) | (/*PG*/1U<<31);
  sregs.cr3 = GPHYS(pagetable1);
  sregs.cr4 = (/*PAE*/1<<5);
  sregs.efer = (/*LME*/1<<8) | (/*LMA*/1<<10);
  struct kvm_regs regs = {
    .rdi = 0x1f00,
    .rsi = 0xf00d,
    .rip = 0x0,
    .rflags = 2
  };

  /*
   * set up segments.
   * note that this never actually writes the segments into a GDT; we just load
   * the segments directly.
   */
  sregs.cs = (struct kvm_segment) {
    .limit = 0xffffffff,
    .selector = (1 << 3) | 3,
    .present = 1,
    .type = 11/*Code: Execute/Read, accessed*/,
    .dpl = 3,
    .s = 1/*code or data segment*/,
    .l = 1/*64-bit code segment, must have .db clear*/,
    .g = 1/*segment limit interpreted in 4KiB units*/
  };
  sregs.ds = sregs.cs;
  //sregs.ds.l = 0;
  sregs.ds.type = 3/*Data: Read/Write, accessed*/;
  sregs.ds.selector = (2 << 3) | 3;
  sregs.es = sregs.fs = sregs.gs = sregs.ss = sregs.ds;

  if (ioctl(vcpu, KVM_SET_SREGS, &sregs))
    err(1, "set sregs");
  if (ioctl(vcpu, KVM_SET_REGS, &regs))
    err(1, "set regs");
  gmem.steal_time.version = 1;
  gmem.steal_time.preempted = 0xaa;
  wrmsr(vcpu, MSR_KVM_STEAL_TIME, 1 | GPHYS(steal_time));

  unsigned char *vmmcall = gmem.guest_code + (guest_code_vmmcall - guest_code_begin);
  printf("VMMCALL bytes: %02hhx %02hhx %02hhx\n", vmmcall[0], vmmcall[1], vmmcall[2]);

  pthread_t thread;
  if (pthread_create(&thread, NULL, thread_fn, NULL))
    errx(1, "pthread_create");

  if (ioctl(vcpu, KVM_RUN, 0))
    err(1, "run vcpu");
  printf("exit_reason = %d\n", vcpu_state->exit_reason);
  if (vcpu_state->exit_reason == KVM_EXIT_FAIL_ENTRY) {
    printf("KVM_EXIT_FAIL_ENTRY happened: hardware_entry_failure_reason = 0x%lx (see arch/x86/include/uapi/asm/vmx.h)\n",
           (unsigned long)vcpu_state->fail_entry.hardware_entry_failure_reason);
  }
  if (vcpu_state->exit_reason == KVM_EXIT_SHUTDOWN)
    printf("KVM_EXIT_SHUTDOWN happened\n");
  if (ioctl(vcpu, KVM_GET_REGS, &regs))
    err(1, "get regs");
  printf("guest RIP: 0x%llx\n", regs.rip);
  printf("VMMCALL bytes: %02hhx %02hhx %02hhx\n", vmmcall[0], vmmcall[1], vmmcall[2]);
  printf("kvm_steal_time: version=0x%x preempted=0x%hhx\n", gmem.steal_time.version, gmem.steal_time.preempted);
}


